local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")
local playerName = "cta3_player0"
local racerNames = {"cta3_ai0", "cta3_ai1", "cta3_ai2"}
local trackerNames = {"cta3_marker0", "cta3_marker1", "cta3_marker2"}
local messages = {"You busted all of them! Time to make them pay for their crimes.", "You busted all of them! That was quick, they're terrified of you! Excellent.", "You busted all of them! It was a tedious task, but your patience got you the win.", "One of them escaped. Fail!"}

local racerAggro = 1.1 --Main racer aggression
local farDist = 300 --Used for distance based AI aggression mode
local readyDist = 150 --Used for AI trackers and chase logic
local nearDist = 30 --Used for arrest radius
local crawlSpeed = 3 --Slow speed
local zShift = 15 --Shifts tracker beacons up
local alpha = 0 --Alpha transparency
local highlight = true --Tracker mode
local mainTimer = 0
local overTime = 0

local player = {}
local racers = {}
local Racer = {}

function Racer:new(obj, pos, tPos, vel, state, arrestTimer, arrestRadius, evadeTimer, tracker)
	local self = setmetatable({}, Racer)
	self.__index = self
	self.obj = obj or nil
	self.pos = pos or nil
	self.tPos = tPos or nil
	self.vel = vel or nil
	self.state = state or nil
	self.arrestTimer = arrestTimer or 0
	self.arrestRadius = arrestRadius or 0
	self.evadeTimer = evadeTimer or 0
	self.tracker = tracker or nil
	return self
end

local function arrestRacer(aRacer)
	if aRacer.state ~= "stop" then
		if aRacer.arrestTimer % 1 == 0 then --Countdown happens
			local countdown = 5 - math.modf(aRacer.arrestTimer)
			if countdown > 0 then
				helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
			end
		end
		aRacer.arrestTimer = aRacer.arrestTimer + 0.25
	end 	
	if aRacer.arrestTimer > 5 then
		helper.flashUiMessage("Busted!", 2, true)
		helper.setAiMode(aRacer.obj.name, "stopping")
		aRacer.arrestTimer = 0
		aRacer.state = "stop"
		player.score = player.score + 1
	end
end

local function onScenarioUIReady(state)
	player = {obj = scenetree.findObject(playerName), pos = nil, tPos = nil, vel = nil, dist = nil, tDist = nil, score = 0, cheater = false}
	for i, v in ipairs(racerNames) do
		racers[i] = Racer:new(scenetree.findObject(racerNames[i]), nil, nil, nil, "run", 0, nearDist, 0, scenetree.findObject(trackerNames[i]))
		racers[i].tracker.hidden = true
		racers[i].tracker:setScale(Point3F(6, 6, 60))
		racers[i].tracker.instanceColor = ColorF(1, 0, 0, alpha):asLinear4F()
	end
	mainTimer = 0
	overTime = 0
end

local function onRaceStart()
	for i, v in ipairs(racers) do
		racers[i].tracker.hidden = false
		helper.setAiMode(racers[i].obj.name, "random")
	end
	player.obj:queueLuaCommand("obj:queueGameEngineLua('value = ('..obj:calcBeamStats().total_weight..')')")
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	for i, v in ipairs(racers) do
		racers[i].pos = racers[i].obj:getPosition()
		racers[i].vel = map.objects[map.objectNames[racers[i].obj.name]].vel:length()
		player.dist = (player.pos - racers[i].pos):len()
		if i == mainTimer - 7 then
			print(value)
			helper.setAiMode(racers[i].obj.name, "flee")
		end
		if racers[i].state ~= "stop" then
			if player.dist >= farDist then
				helper.setAiAggression(racers[i].obj.name, racerAggro / 1.25)
			else
				helper.setAiAggression(racers[i].obj.name, racerAggro)
			end
			if player.dist <= readyDist then
				if racers[i].vel <= crawlSpeed then 
					racers[i].arrestRadius = racers[i].arrestRadius + 1 --Arrestable distance grows by 4 metres per second to prevent rare but unfair situations
				else
					racers[i].arrestRadius = nearDist
				end
				if mainTimer >= 5 and player.dist <= racers[i].arrestRadius and racers[i].vel <= crawlSpeed then
					arrestRacer(racers[i])
				else
					racers[i].arrestTimer = 0
				end
			end	
		end
	end
	if player.score >= #racers then
		if overTime == 0 then
			overTime = mainTimer + 1 --1 second delay at the end
		end
		if mainTimer >= overTime then
			if mainTimer <= 300 then
				scenario.finish({msg = messages[2]})
			elseif mainTimer >= 900 then
				scenario.finish({msg = messages[3]})
			else
				scenario.finish({msg = messages[1]})
			end
		end
	end
	if commands.isFreeCamera(0) and player.cheater == false then
		player.obj:setPosition(vec3(0, 0, 1200):toPoint3F())
		player.cheater = true
	end
end

local function onPreRender(dt, dtSim)
	if highlight == true and mainTimer > 0 then
		player.tPos = player.obj:getPosition()
		for i, v in ipairs(racers) do
			racers[i].tPos = racers[i].obj:getPosition()
			player.tDist = (player.tPos - racers[i].tPos):len()
			alpha = math.min(0.7, math.max(0, (player.tDist - 250) / 350)) --Fade from 250 to 500
			racers[i].tracker:setPosition(vec3(racers[i].tPos.x, racers[i].tPos.y, racers[i].tPos.z + zShift):toPoint3F())
			if racers[i].state == "run" then
				racers[i].tracker.instanceColor = ColorF(1, 0, 0, alpha):asLinear4F()
			elseif racers[i].state == "stop" then
				racers[i].tracker.hidden = true
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onPreRender = onPreRender

return M